/*
 * Copyright (c) 2002, Jason Tishler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * A copy of the GNU General Public License can be found at
 * http://www.gnu.org/
 *
 * Written by Jason Tishler <jason@tishler.net>
 */

#include <stdlib.h>
#include <iostream>
#include <string>
#include "RebaseConfigParser.h"
#include "RebaseMemoryBuilder.h"

RebaseConfigParser::RebaseConfigParser (const string &aFileName) :
  theStream (aFileName.c_str ())
{
  if (!theStream)
    {
      cerr << "Cannot open config file " + aFileName + " for reading" << endl;
      // FIXME: throw exception?
    }
}

RebaseConfigParser::~RebaseConfigParser ()
{
}

bool
RebaseConfigParser::parse(RebaseBuilder &aBuilder)
{
  bool status = true;

  while (theStream)
    {
      string aHeader = getHeader ();
      if (aHeader == theVersionHeader)
	{
	  aBuilder.buildVersion ();
	  status = parseVersion (aBuilder);
	}
      else if (aHeader == theFreeListHeader)
	{
	  aBuilder.buildFreeList ();
	  status = parseFreeList (aBuilder);
	}
      else if (aHeader == theUsedListHeader)
	{
	  aBuilder.buildUsedList ();
	  status = parseUsedList (aBuilder);
	}
      else if (aHeader.size () > 0)
	{
	  status = false;
	}

      if (!status)
	{
	  cerr << "invalid configuration stream" << endl;
	  break;
	}
    }

  return status;
}

bool
RebaseConfigParser::parseVersion (RebaseBuilder &aBuilder)
{
  string aVersion = getLine (theStream);
  aBuilder.buildVersionEntry (aVersion);
}

bool
RebaseConfigParser::parseFreeList (RebaseBuilder &aBuilder)
{
  const unsigned int aNumFields = 2;

  for (string aLine; (aLine = getLine (theStream)).size () > 0;)
    {
      Fields aFields;
      getFields (aLine, aFields);
      if (aFields.size () != aNumFields)
	continue;
      unsigned long aBegin = strtoul (aFields[0].c_str (), 0, 0);
      unsigned long anEnd = strtoul (aFields[1].c_str (), 0, 0);
      aBuilder.buildFreeListEntry (aBegin, anEnd);
    }
}

bool
RebaseConfigParser::parseUsedList (RebaseBuilder &aBuilder)
{
  const unsigned int aNumFields = 3;

  for (string aLine; (aLine = getLine (theStream)).size () > 0;)
    {
      Fields aFields;
      getFields (aLine, aFields);
      if (aFields.size () != aNumFields)
	continue;
      string aPathname = aFields[0];
      unsigned long aBase = strtoul (aFields[1].c_str (), 0, 0);
      unsigned long aSize = strtoul (aFields[2].c_str (), 0, 0);
      aBuilder.buildUsedListEntry (aPathname, aBase, aSize);
    }
}

void
RebaseConfigParser::getFields (const string &aString, Fields &aFields)
{
  const string aFieldSep = "\t";

  for (unsigned int aBegin = 0, aSize = aString.size (); aBegin < aSize;)
    {
      unsigned int anEnd = aString.find (aFieldSep, aBegin);
      if (anEnd == string::npos)
	anEnd = aSize + 1;
      string aSubString = aString.substr (aBegin, anEnd - aBegin);
      aFields.push_back (aSubString);
      aBegin = anEnd + 1;
    }
}

string
RebaseConfigParser::getHeader ()
{
  // Eat lines until we find a section header
  string aHeader;
  for (string aLine; ::getline (theStream, aLine);)
    {
      unsigned int size = aLine.size();
      if (aLine[0] == '[' && aLine[size - 1] == ']')
	{
	  aHeader = aLine.substr (1, size - 2);
	  break;
	}
    }

  return aHeader;
}

string
RebaseConfigParser::getLine (istream &aStream)
{
  string aLine;
  for (string aBuffer; aStream;)
    {
      char aChar = aStream.peek ();
      if (aChar == '[')
	break;
      ::getline (aStream, aBuffer);
      if (aStream && !isSpaceLine (aBuffer))
	{
	  aLine = aBuffer;
	  break;
	}
    }

  return aLine;
}

bool
RebaseConfigParser::isSpaceLine (const string &aLine)
{
  bool status = true;

  for (unsigned int i = 0, size = aLine.size (); i < size; i++)
    {
      if (!::isspace (aLine[i]))
	{
	  status = false;
	  break;
	}
    }

  return status;
}

const string RebaseConfigParser::theVersionHeader = "Version";
const string RebaseConfigParser::theFreeListHeader = "Free List";
const string RebaseConfigParser::theUsedListHeader = "Used List";
